from django.contrib import admin
from django.utils.html import format_html


class MetaClass(admin.ModelAdmin.__class__):
    def __new__(mcs, name, bases, attrs):
        new_class = super(MetaClass, mcs).__new__(mcs, name, bases, attrs)
        if new_class.model_cls:
            admin.site.register(new_class.model_cls, new_class)
        return new_class


class ModelAdminAbstract(admin.ModelAdmin, metaclass=MetaClass):
    model_cls = None


class ModelAdminBase(ModelAdminAbstract):
    img_width = 50

    def format_img(self, img):
        if img:
            return format_html('<img src="{0}" width="{1}" height="{1}"/>'.format(img, self.img_width))
        else:
            return ""

    def format_a_img(self, img):
        if img:
            return format_html('<a href="{0}" target="_blank"><img src="{0}" width="{1}" height="{1}"/></a>'.
                               format(img, self.img_width))
        else:
            return ""

    @staticmethod
    def format_a(url, text):
        if url:
            return format_html("<a href='{url}'target='_blank'>{text}</a>", url=url, text=text)
        return ""

    def format_img_list(self, img_list):
        return format_html("".join([self.format_img(x) for x in img_list]))

    def format_a_img_list(self, img_list):
        return format_html("".join([self.format_a_img(x) for x in img_list]))


def _auto_list_display(model_cls, up_limit=None):
    result = []
    primary_key = None
    for field in model_cls._meta.get_fields():
        if getattr(field, 'primary_key', False):
            primary_key = field.name
        elif field.is_relation:
            if field.many_to_one:
                result.append(field.name)
        else:
            result.append(field.name)
    if primary_key:
        result.insert(0, primary_key)
    if up_limit:
        result = result[:up_limit]
    return tuple(result)


def admin_register(model_cls, list_display=None):
    type(
        'admin',
        (ModelAdminBase,),
        {'model_cls': model_cls, 'list_display': list_display or _auto_list_display(model_cls)}
    )


def admin_register_batch(model_cls_list):
    for model_cls in model_cls_list:
        admin_register(model_cls)



from dektools.encode.b62 import Base62Int
from .snowflake import SnowflakeGenerator


class UnionId:
    charset_unordered = 'ozRMrxcwFbTpOLnq4CgAeBI1JNk3XWHf0Dms9YhydSviVauZ8UGjPlE572K6tQ'

    def __init__(self, unordered=False, max_length=None):
        self.unordered = unordered
        self.max_length = max_length
        self.snowflake_generator = SnowflakeGenerator.new_instance()
        self.base62 = Base62Int(
            self.snowflake_generator.MAX_VALUE,
            self.charset_unordered if unordered else Base62Int.charset_default
        )

    def new_id(self):
        uid = self.base62.to_str_extend(self.snowflake_generator.get_next_id(), self.length)
        if self.unordered:
            s1 = uid[:self.base62.max_length_int]
            s2 = uid[self.base62.max_length_int:]
            if len(s1) > len(s2):
                s2, s1 = s1, s2
            m = len(s1)
            r = ''
            for i, x in enumerate(s2):
                r += x + (s1[i] if i < m else '')
            return r
        return uid

    @property
    def length(self):
        return self.base62.get_max_length(self.max_length)



from django.core.management import call_command
from ..base import CommandBasic


class Command(CommandBasic):
    def handle(self):
        call_command('mediarestore', interactive=False)
        call_command('dbrestore', interactive=False)



import os
from django.core.management import call_command
from dektools.file import clear_dir
from ..base import CommandBasic
from .utils import get_backup_location


class Command(CommandBasic):
    def handle(self):
        location = get_backup_location()
        if location and os.path.isdir(location):
            clear_dir(location)
        call_command('dbbackup', interactive=False)
        call_command('mediabackup', interactive=False)



import os
from django.conf import settings
from ..base import CommandBasic
from .utils import get_backup_location


class Command(CommandBasic):
    def handle(self):
        location = get_backup_location()
        if location:
            os.makedirs(location)
        os.makedirs(settings.MEDIA_ROOT)



import os
from pathlib import Path
from sysconfig import get_paths
from importlib import import_module
from dektools.dict import assign
from dektools.yaml import yaml
from dektools.module import get_module_attr


django_module = os.environ['DJANGO_SETTINGS_MODULE'].rpartition('.')[0]


INSTALLED_APPS = get_module_attr(f"{django_module}.settings_.INSTALLED_APPS")

SETTINGS_FILE_NAME = 'settings.app.yaml'

PATH_LIB = Path(get_paths()['platlib'])
PATH_BASE = Path(import_module(django_module).__file__).resolve().parent.parent

INSTALLED_APPS_REVERSED = list(reversed(INSTALLED_APPS))


def load_settings_app(app_name, path):
    config = {}
    config_first = {}
    for app in INSTALLED_APPS_REVERSED:
        app_path = app.replace('.', os.sep)
        for path_dir in (PATH_BASE, PATH_LIB):
            p = path_dir / app_path / SETTINGS_FILE_NAME
            if os.path.isfile(p):
                data = yaml.load(p)
                data_app = data.get(app_name) or {}
                data_path = data_app.get(path)
                if data_path is not None:
                    if app_name == app:
                        config_first = data_path
                    else:
                        config = assign(config, data_path)
    config = assign(config_first, config)
    return config



import os
from django.db import models
from ..snowflake import SnowflakeGenerator


class ImageField(models.ImageField):
    def __init__(self, *args, **kwargs):
        self.snowflake_generator = SnowflakeGenerator.new_instance()
        super().__init__(*args, **kwargs)

    def generate_filename(self, instance, filename):
        filename = str(self.snowflake_generator.get_next_id()) + os.path.splitext(filename)[-1]
        return super().generate_filename(instance, filename)



import re
from django_extensions.management.commands.shell_plus import Command
from dektools.file.operation import read_text
from dekshell.core.markers.base import MarkerBase, MarkerWithEnd


def django_setup(env):
    if not env.context.get('__django__'):
        import sys, os
        sys.path[:] = sys.path[:] + [os.getcwd()]

        project_name = re.search(
            r"""os.environ.setdefault\(['"]{1}DJANGO_SETTINGS_MODULE['"]{1}, ['"]{1}([0-9a-zA-Z_]+).settings['"]{1}\)""",
            read_text('./manage.py')
        ).groups()[0]

        import os, django
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", f"{project_name}.settings")
        django.setup()
        env.context['__django__'] = True


class DjangoMarker(MarkerBase):
    tag_head = "@django"

    def exec(self, env, command, marker_node, marker_set):
        args = self.split_raw(command, 1)
        django_setup(env)

        vars_plus = Command().get_imported_objects(dict(quiet_load=True))
        vars_extra = {
            'User': vars_plus['get_user_model']()
        }

        self.eval(env, args[1], vars_plus | vars_extra)


class DjangoBlockMarker(MarkerWithEnd):
    tag_head = "@django-block"

    def exec(self, env, command, marker_node, marker_set):
        args = self.split_raw(command, 1)
        django_setup(env)

        vars_plus = Command().get_imported_objects(dict(quiet_load=True))
        vars_extra = {
            'User': vars_plus['get_user_model']()
        }

        self.eval_multi(
            env,
            '\n'.join([child.command for child in marker_node.children]),
            vars_plus | vars_extra
        )

        return []



import json
from django.db.models import Model
from django.forms.models import model_to_dict


def model_to_json(model, **kwargs):
    return json.dumps(model_to_data(model), ensure_ascii=False, **kwargs)


def model_to_data(model):
    return sure_model_data(model_to_dict(model))


def sure_model_data(data):
    if isinstance(data, dict):
        for k in list(data):
            v = data[k]
            if isinstance(v, Model):
                data[k] = model_to_data(data[k])
            elif isinstance(v, (dict, list)):
                data[k] = sure_model_data(data[k])
    elif isinstance(data, list):
        for index, item in enumerate(data):
            if isinstance(item, Model):
                data[index] = model_to_data(item)
            elif isinstance(item, (dict, list)):
                data[index] = sure_model_data(item)
    return data



import os
from django.conf import settings
from django.apps import apps
from dektools.file import normal_path

project_dir = normal_path(settings.BASE_DIR)

project_dir_prefix = project_dir + os.path.sep


def list_migrations():
    result = set()
    for app in apps.get_app_configs():
        app_path = app.path
        if app_path.startswith(project_dir_prefix):
            migrations_path = os.path.join(app_path, 'migrations')
            if os.path.isdir(migrations_path):
                for item in os.listdir(migrations_path):
                    item_path = os.path.join(migrations_path, item)
                    if os.path.isfile(item_path) and item != '__init__.py' and os.path.splitext(item)[-1] == '.py':
                        result.add(item_path)
    return project_dir, result



# File | Settings | Build, Execution, Deployment | Console | Django Console
# Starting script

from django_extensions.management.commands.shell_plus import Command;globals().update(Command().get_imported_objects({}))



import importlib
import json
import types
from django.apps import apps
from django.urls import reverse, NoReverseMatch, get_resolver
from django.views.generic import View
from django.core.files.uploadedfile import UploadedFile


class ReqExecute:
    EMPTY_VALUE = object()

    def __init__(self, client, method, url, kwargs):
        self.client = client
        self.method = method
        self.url = url
        self.kwargs = kwargs

    @property
    def formatted_kwargs(self):
        kwargs = self.kwargs.copy()
        data = kwargs.pop('json', self.EMPTY_VALUE)
        if data is not self.EMPTY_VALUE:
            if self.method == 'get':
                kwargs.update(dict(
                    data={k: v for k, v in data.items() if v is not None}
                ))
            else:
                if self.has_file(data):
                    kwargs.update(dict(data=data))
                else:
                    kwargs.update(dict(
                        data=json.dumps(data),
                        content_type='application/json'
                    ))
        return kwargs

    def get_rsp(self):
        return getattr(self.client, self.method)(self.url, **self.formatted_kwargs)

    @staticmethod
    def has_file(data):
        for k, v in data.items():
            if isinstance(v, list) and v and isinstance(v[0], UploadedFile):
                return True
            elif isinstance(v, UploadedFile):
                return True
        return False

    @staticmethod
    def try_attach(rsp):
        try:
            json_string = rsp.content.decode('utf-8')
            try:
                rsp.json = json.loads(json_string)
            except json.decoder.JSONDecodeError:
                pass
        except UnicodeDecodeError:
            pass
        return rsp

    def run(self):
        return self.try_attach(self.get_rsp())


class ReqHeaders:
    def __init__(self):
        self.headers = {}

    def update(self, data):
        self.headers.update(data)

    def set(self, data):
        self.headers = {**data}

    def format(self):
        result = []
        for k, v in self.headers.items():
            if isinstance(k, str):
                k = k.encode('utf-8')
            if isinstance(v, str):
                v = v.encode('utf-8')
            result.append((k, v))
        return result


class ReqTool:
    full_http_method_names = View.http_method_names
    _auth_key = ''

    @classmethod
    def set_auth_key(cls, auth_key):
        cls._auth_key = auth_key

    @classmethod
    def req(cls, client, action, method_name=None, reverse_kwargs=None, **kwargs):
        url, methods = cls.get_action_info(action, **reverse_kwargs or {})
        if method_name is None:
            method_name = methods[0]
        headers = ReqHeaders()
        headers.update(kwargs.pop('headers', {}))
        auth_kwargs = {}
        if cls._auth_key:
            auth_kwargs.update(dict(HTTP_AUTHORIZATION=f'Token {cls._auth_key}'))
        return ReqExecute(client, method_name, url, {**dict(headers=headers.format(), **auth_kwargs), **kwargs}).run()

    @classmethod
    def get_url(cls, action, **kwargs):
        return cls.get_action_info(action, **kwargs)[0]

    _view2query_cache = {}

    @classmethod
    def view2query(cls, action):
        result = cls._view2query_cache.get(action)
        if result:
            return result
        for ur in get_resolver().url_patterns:
            rd = ur.reverse_dict
            ns = ur.namespace
            if action in rd:
                for k, v in rd.items():
                    if k != action and v == rd[action]:
                        if ns:
                            result = f'{ns}:{k}'
                        else:
                            result = k
                        cls._view2query_cache[action] = result
                        return result

    @classmethod
    def get_action_info(cls, action, **kwargs):
        try:
            action_query = action
            if not isinstance(action, str):
                action_query = cls.view2query(action_query) or action_query
            url = reverse(action_query, **kwargs)
            return url, cls.sort_methods(action.view_class.http_method_names)
        except NoReverseMatch:
            pass
        for (ns, basename), (view_set, routes) in cls.get_view_sets().items():
            if getattr(view_set, action.__name__, None) is action:
                for route_name, mapping in routes:
                    url_name = getattr(action, "url_name", action.__name__)
                    http_method_names = mapping.get(url_name)
                    if http_method_names:
                        n = route_name.format(basename=basename)
                        if ns:
                            n = f'{ns}:{n}'
                        return reverse(n, **kwargs), cls.sort_methods(http_method_names)
        raise NoReverseMatch(f"Can not find the action[{action}]")

    @classmethod
    def sort_methods(cls, array):
        return sorted(array, key=lambda x: cls.full_http_method_names.index(x))

    @classmethod
    def get_view_sets(cls):
        def result():
            ns_map = {}
            for ur in get_resolver().url_patterns:
                if isinstance(ur.urlconf_module, types.ModuleType):
                    ns_map[ur.urlconf_module] = ur.namespace

            r = {}
            for app in apps.get_app_configs():
                try:
                    urls = importlib.import_module(f"{app.name}.urls")
                    router = getattr(urls, "router", None)
                    if router:
                        ns = ns_map[urls] or None
                        for _, viewset, basename in router.registry:
                            routes = router.get_routes(viewset)
                            array = []
                            for route in routes:
                                m = {}
                                for k in sorted(route.mapping):
                                    ay = m.setdefault(route.mapping[k], [])
                                    ay.append(k)
                                array.append((route.name, m))
                            r[(ns, basename)] = viewset, array
                except ModuleNotFoundError:
                    pass
            return r

        if not hasattr(cls, "_get_view_sets"):
            setattr(cls, "_get_view_sets", result())
        return getattr(cls, "_get_view_sets")



import re
from django.utils.translation import ugettext_lazy as _
from django.core.validators import EmailValidator
from django.core.exceptions import ValidationError
from rest_framework import serializers


def is_valid_mobile(s):
    return re.match("^(1[3-9])\\d{9}$", s)


def validate_mobile(s):
    if not is_valid_mobile(s):
        raise serializers.ValidationError(_('手机号格式错误'))


def is_valid_email(s):
    try:
        EmailValidator()(s)
        return True
    except ValidationError:
        return False


def validate_email(s):
    if not is_valid_email(s):
        raise serializers.ValidationError(_('邮箱格式错误'))



from django.core.management import call_command
from ..base import CommandBasic


class Command(CommandBasic):
    just_for_debug_mode = True

    def handle(self):
        call_command('reset_db', interactive=False)
        call_command('djdelmigrations')
        call_command('makemigrations')
        call_command('migrate')



import os
from pathlib import Path
from importlib import import_module

BASE_DIR = Path(import_module(os.environ['DJANGO_SETTINGS_MODULE'].rpartition('.')[0]).__file__).parent.parent

DEKDJTOOLS_SNOWFLAKE_INSTANCE = 0

SILENCED_SYSTEM_CHECKS = ['models.W042']  # dynaconf nearly missing `is_overridden`

REST_FRAMEWORK = {
    'DATETIME_FORMAT': '%Y-%m-%d %H:%M:%S.%f%z',
    'DEFAULT_PERMISSION_CLASSES': ['rest_framework.permissions.DjangoModelPermissionsOrAnonReadOnly'],
}

DBBACKUP_STORAGE = 'django.core.files.storage.FileSystemStorage'
DBBACKUP_STORAGE_OPTIONS = {'location': BASE_DIR / '.dbbackup'}
DBBACKUP_CONNECTOR_MAPPING = {'psqlextra.backend': 'dbbackup.db.postgresql.PgDumpBinaryConnector'}



from django.db import models
from django.apps import apps
from django.utils.functional import cached_property
from django.db.models.fields.mixins import FieldCacheMixin


class RelDataDescriptor:
    def __init__(self, field):
        self.field = field

    def __get__(self, instance, cls=None):
        return self.field.get_cached_value(instance, default=None)

    def __set__(self, instance, value):
        if value is not None:
            if isinstance(value, self.field.custom_related_model):
                self.field.set_cached_value(instance, value.format_as_data())
            else:
                raise ValueError(
                    'Cannot assign "%r": "%s.%s" must be a "%s" instance.' % (
                        value,
                        instance._meta.object_name,
                        self.field.name,
                        self.field.custom_related_model._meta.label,
                    )
                )


class RelDataField(FieldCacheMixin, models.JSONField):
    related_accessor_class = RelDataDescriptor

    def __init__(self, to=None, **kwargs):
        self.to = to
        super().__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['to'] = self.custom_related_model._meta.label
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name, private_only=False, **kwargs):
        super().contribute_to_class(cls, name, private_only=private_only)
        setattr(cls, self.name, self.related_accessor_class(self))

    def get_cache_name(self):
        return self.name

    @cached_property
    def custom_related_model(self):
        if isinstance(self.to, str):
            array = self.to.rsplit('.', 1)
            if len(array) == 1:
                app, name = self.model._meta.app_label, array[0]
            else:
                app, name = array
            return apps.get_model(app, name)
        else:
            return self.to



from pymediainfo import MediaInfo


def get_media_metadata(filepath):
    media_info = MediaInfo.parse(filepath)
    tracks = {x['track_type']: x for x in (t.to_data() for t in media_info.tracks)}
    general = tracks['General']
    base_info = dict(s=general['file_size'], e=general.get('file_extension'))
    image = tracks.get('Image')
    if image:
        return dict(
            **base_info,
            f=image['format'],
            w=image['width'],
            h=image['height'],
        )
    video = tracks.get('Video')
    if video:
        return dict(
            **base_info,
            f=video['format'],
            d=video['duration'],
            w=video['width'],
            h=video['height'],
        )
    audio = tracks.get('Audio')
    if audio:
        return dict(
            **base_info,
            f=audio['format'],
            d=audio['duration']
        )
    return base_info



import os
import re
from django.conf import settings
from ..base import CommandBasic


class Command(CommandBasic):
    def handle(self):
        if self.just_for_debug_mode():
            return
        apps = set(settings.INSTALLED_APPS)
        for root, dirs, files in os.walk(settings.BASE_DIR):
            for f in files:
                if root.endswith('migrations') and re.match(r'^\d{4}_[a-z_0-9]+.py$', f):
                    if os.path.dirname(root).replace(str(settings.BASE_DIR), '').strip(os.sep).replace(os.sep,
                                                                                                       '.') in apps:
                        self.stdout.write(f'delete: {os.path.join(root, f)}')
                        os.remove(os.path.join(root, f))
